import numpy as np
from gameanalysis import rsgame


def deviation_payoffs(prof_sched, game, mix, num, *, boots=0, chunk_size=None):
    """Bootstrap deviation payoffs

    Parameters
    ----------
    prof_sched : Scheduler
        The scheduler to sample profiles from.
    game : RsGame
        The game defining what's sampled.
    mix : ndarray
        The mixture to calculate the regret of.
    num : int
        The number of samples to gather.
    boots : int, optional
        The number of bootstrap samples to take. The accuracy of bootstrap is
        independent of this number, but more will reduce the variance of the
        underlying confidence bounds. The default will compute no bootstrap
        gains.
    chunk_size : int, optional
        An implementation detail specifying how frequently profiles are
        scheduled since this algorithm inherently operates in a streaming
        manner. Ideally this number should be set such that the time to
        schedule and process chunk_size roughly equals the time for one
        simulation. It also controls how much memory this uses. By default this
        is set to ten times the number of bootstraps, or 1000 if no bootstraps
        are requested.

    Notes
    -----
    This uses memory on the order of `boots + chunk_size`. It is inefficient if
    `num` is less than boots.

    Returns
    -------
    mean_gains : ndarray (num_strats,)
        The mean deviation payoffs from the mixture.
    boot_gains : ndarray (boots, num_strats)
        The deviation payoffs for each bootstrap sample.
    """
    chunk_size = chunk_size or boots * 10 or 1000
    game = rsgame.emptygame_copy(game)
    profiles = _chunk_profiles(prof_sched, game, mix, num, chunk_size)
    devs = np.empty(game.num_strats)
    mean_devs = np.zeros(game.num_strats)
    boot_devs = np.zeros((boots, game.num_strats))
    remaining = np.empty(boots, int)
    remaining.fill(num)
    for i in range(num):
        for j in range(game.num_strats):
            devs[j] = next(profiles)[j]
        mean_devs += (devs - mean_devs) / (i + 1)
        samps = np.random.binomial(remaining, 1 / (num - i))
        remaining -= samps
        boot_devs += samps[:, None] * devs / num

    return mean_devs, boot_devs


def _chunk_profiles(sched, game, mix, num, chunk_size):
    """Return a generator over payoffs that schedules somewhat efficiently"""
    proms = []
    while 0 < num:
        new_profs = game.random_deviator_profiles(
            mix, min(num, chunk_size)).reshape((-1, game.num_strats))
        num -= chunk_size
        new_proms = [sched.schedule(prof) for prof in new_profs]
        for prom in proms:
            yield prom.get()
        proms = new_proms
    for prom in proms:
        yield prom.get()



import json

import numpy as np
from gameanalysis import nash
from gameanalysis import paygame
from gameanalysis import rsgame
from gameanalysis.reduction import deviation_preserving as dpr
from gameanalysis.reduction import identity as ir


def add_parser(subparsers):
    parser = subparsers.add_parser(
        'brute', help="""Compute equilibria by sampling all profiles""",
        description="""Samples profiles from the entire game, and then runs
        standard equilibrium finding. For games with a large number of players,
        a reduction should be specified.""")
    parser.add_argument(
        '--dpr', metavar='<role:count,role:count,...>', help="""Specify a DPR
        reduction.""")
    parser.add_argument(
        '--regret-thresh', metavar='<reg>', type=float, default=1e-3,
        help="""Regret threshold for a mixture to be considered an equilibrium.
        (default: %(default)g)""")
    parser.add_argument(
        '--dist-thresh', metavar='<norm>', type=float, default=1e-3,
        help="""Norm threshold for two mixtures to be considered distinct.
        (default: %(default)g)""")
    return parser


def parse_reduction(red):
    strats = (s.strip().split(':') for s in red.strip().split(','))
    return {r.strip(): int(c) for r, c in strats}


def run(scheduler, game, args):
    game = rsgame.emptygame_copy(game)
    red = ir
    red_players = None
    if args.dpr is not None:
        red_players = game.from_role_json(parse_reduction(args.dpr), dtype=int)
        red = dpr
    redgame = red.reduce_game(game, red_players)
    profiles = red.expand_profiles(game, redgame.all_profiles())
    promises = [scheduler.schedule(prof) for prof in profiles]
    payoffs = np.concatenate([prom.get()[None] for prom in promises])
    redgame = red.reduce_game(paygame.game_replace(
        game, profiles, payoffs), red_players)
    eqa = nash.mixed_nash(redgame, regret_thresh=args.regret_thresh,
                          dist_thresh=args.dist_thresh)
    for eqm in eqa:
        json.dump(game.to_mix_json(eqm), args.output)
        args.output.write('\n')



import argparse
import json

import numpy as np
from gameanalysis import rsgame

from egta import bootstrap


def add_parser(subparsers):
    parser = subparsers.add_parser(
        'bootstrap', aliases=['boot'], help="""Compute the regret and surplus
        of a mixture""", description="""Samples profiles to compute a sample
        regret and sample surplus of the mixture. By optionally specifying
        percentiles, bootstrap confidence bounds will be returned.""")
    parser.add_argument(
        'mixture', metavar='<mixture-file>', type=argparse.FileType('r'),
        help="""A file with the json formatted mixture to compute the regret
        of.""")
    parser.add_argument(
        'num', metavar='<num-samples>', type=int, help="""The number of samples
        to compute for the regret. One sample for each strategy in the game
        will be taken.""")
    parser.add_argument(
        '--percentiles', '-p', metavar='<percentile>,...', default=(),
        type=lambda s: list(map(float, s.split(','))), help="""A comma
        separated list of confidence percentiles to compute. Specifying
        automatically switches this to use a bootstrap method which takes
        memory proportional to the number of bootstrap samples to run.""")
    parser.add_argument(
        '--boots', '-b', metavar='<num-samples>', type=int, default=101,
        help="""The number of bootstrap samples to take if percentiles is
        specified. More samples will produce less variable confidence bounds at
        the cost of more time. (default: %(default)d)""")
    parser.add_argument(
        '--chunk-size', '-c', metavar='<chunk-size>', type=int, help="""Roughly
        the number of profiles to be scheduled simultaneously. This will be set
        to 10 * boots or 1000 if unspecified. The time to run a simulation
        should roughly correspond to the time to process chunk_size payoffs.
        Keeping this on the order of the number of bootstraps keeps the memory
        requirement constant.""")
    return parser


def run(scheduler, game, args):
    game = rsgame.emptygame_copy(game)
    if not args.percentiles:
        args.boots = 0
    mix = game.from_mix_json(json.load(args.mixture))
    means, boots = bootstrap.deviation_payoffs(
        scheduler, game, mix, args.num, boots=args.boots,
        chunk_size=args.chunk_size)

    exp_means = np.add.reduceat(means * mix, game.role_starts)
    exp_boots = np.add.reduceat(boots * mix, game.role_starts, 1)

    gain_means = means - exp_means.repeat(game.num_role_strats)
    gain_boots = boots - exp_boots.repeat(game.num_role_strats, 1)

    reg_means = np.maximum(np.max(gain_means), 0)
    reg_boots = np.maximum(np.max(gain_boots, 1), 0)

    surp_means = exp_means.dot(game.num_role_players)
    surp_boots = exp_boots.dot(game.num_role_players)

    json.dump({
        'surplus': dict(zip(('{:g}'.format(p) for p in args.percentiles),
                            np.percentile(surp_boots, args.percentiles)),
                        mean=surp_means),
        'regret': dict(zip(('{:g}'.format(p) for p in args.percentiles),
                           np.percentile(reg_boots, args.percentiles)),
                       mean=reg_means),
        # TODO this is a bad interface and should probably be somewhere else
        'gains': dict(zip(('{:g}'.format(p) for p in args.percentiles),
                          (game.to_payoff_json(g) for g
                           in np.percentile(gain_boots, args.percentiles, 0))),
                      mean=game.to_payoff_json(gain_means)),
    }, args.output)
    args.output.write('\n')



from egta import profsched


class CountScheduler(profsched.Scheduler):
    """A scheduler that wraps each profile in the mean of n profiles

    Parameters
    ----------
    sched : Scheduler
        The base scheduler that generates payoffs.
    count : int > 0
        The number of times a scheduler in the base scheduler should be sampled
        for each payoff returned by this scheduler.
    """

    def __init__(self, sched, count):
        assert count > 0, "count must be positive {:d}".format(count)
        self._sched = sched
        self._count = count

    def schedule(self, profile):
        return _CountPromise(iter([self._sched.schedule(profile)
                                   for _ in range(self._count)]))

    def __enter__(self):
        self._sched.__enter__()
        return self

    def __exit__(self, *args):
        self._sched.__exit__(*args)


class _CountPromise(profsched.Promise):
    def __init__(self, proms):
        self._proms = proms
        self._value = None

    def get(self):
        if self._value is None:
            # This is doing a streaming mean computation on all of the promises
            # XXX We pull this first instead of initializing zeros because we
            # don't know the size
            count = 1
            self._value = next(self._proms).get().copy()
            for prom in self._proms:
                count += 1
                self._value += (prom.get() - self._value) / count
            self._value.setflags(write=False)
        return self._value



import argparse
import json
import logging
import sys

from egtaonline import api

from egta import eosched


_log = logging.getLogger(__name__)


def add_parser(subparsers):
    parser = subparsers.add_parser(
        'eo', help="""Schedule simulations through egta online""",
        description="""Sample profiles from egta online.""")
    parser.add_argument(
        'sim_memory', metavar='sim-memory-mb', type=int, help="""Maximum memory
        needed to run simulation. Standard limit per core is 4096, but if your
        simulation needs less, you should request less. Note, that if you set
        it too low you'll get biased samples for simulations that require less
        memory to run which could be very bad.""")
    parser.add_argument(
        'sim_time', metavar='sim-time-sec', type=int, help="""Maximum amount of
        time needed to run a single simulation.  Longer times will result in
        simulations taking longer to schedule, but shorter times could result
        in jobs being canceled from going over time which will bias payoff
        results for shorter running simulations.""")
    parser.add_argument(
        '--conf', '-c', metavar='<conf-json>', type=argparse.FileType('r'),
        help="""The configuration file for the simulator. This only needs to be
        specified if a configuration couldn't be found in the specified
        game.""")
    parser.add_argument(
        '--sim-id', metavar='<sim-id>', type=int, help="""The id of the
        simulator to use. If a game id was specified the simulator id will be
        looked up, but it will take a bit longer.""")
    parser.add_argument(
        '--sleep', '-s', metavar='<sleep-seconds>', default=600, type=int,
        help="""Amount of time to sleep in seconds while waiting for output
        from the simulator. Too long and the script will be paused waiting for
        a response from the simulator, too short and a lof of cpu cycles will
        be wasted checking an empty buffer. (default: %(default)d)""")
    parser.add_argument(
        '--max-schedule', metavar='<observations>', default=100, type=int,
        help="""The maximum number of observations to schedule simultaneously.
        This isn't a strict limit, but it won't be violated by much. (default:
        %(default)d)""")
    return parser


def create_scheduler(game, args, configuration=None, simname=None, **_):
    assert configuration is not None or args.conf is not None, \
        "`conf` must be specified or supplied in the game"
    assert simname is not None or args.sim_id is not None, \
        "`sim-id` must be specified or supplied in the game"
    if args.conf is not None:
        configuration = json.load(args.conf)

    if args.sim_id is None:
        with api.EgtaOnlineApi(auth_token=args.auth_string) as ea:
            args.sim_id = next((
                s['id'] for s in ea.get_simulators()
                if '{}-{}'.format(s['name'], s['version']) == simname), None)
    if args.sim_id is None:  # pragma: no cover
        _log.critical("couldn't find simulator with name \"{}\"".format(
            simname))
        sys.exit(1)

    egta = api.EgtaOnlineApi(auth_token=args.auth_string)
    return ApiWrapper(
        egta, args.sim_id, game, args.count, configuration, args.sleep,
        args.max_schedule, args.sim_memory, args.sim_time, args.game_id)


class ApiWrapper(eosched.EgtaOnlineScheduler):
    def __init__(self, api, *args, **kwargs):
        super().__init__(api, *args, **kwargs)
        self._enter_api = api

    def __enter__(self):
        self._enter_api.__enter__()
        return super().__enter__()

    def __exit__(self, *args, **kwargs):
        super().__exit__(*args, **kwargs)
        self._enter_api.__exit__(*args, **kwargs)



import collections
import logging
import queue
import threading
import time
import warnings

import numpy as np
from gameanalysis import collect
from gameanalysis import nash
from gameanalysis import paygame
from gameanalysis import regret
from gameanalysis import rsgame
from gameanalysis import subgame
from gameanalysis import utils
from gameanalysis.reduction import identity as idr


_log = logging.getLogger(__name__)


# FIXME In some cases some threads get left around and this doesn't terminate /
# clean up appropriately.
def inner_loop(prof_sched, game, red=idr, red_players=None, *,
               regret_thresh=1e-3, dist_thresh=1e-3, max_resamples=10,
               subgame_size=3, num_equilibria=1, num_backups=1,
               devs_by_role=False, at_least_one=False):
    """Inner loop a game using a scheduler

    Parameters
    ----------
    prof_sched : Scheduler
        The scheduler used to generate payoff data for profiles.
    game : RsGame
        The gameanalysis basegame for the game to find equilibria on.
    red : reduction, optional
        The reduction to apply to the game before scheduling. Using a reduction
        will sample fewer profiles, while only approximating equilibria. This
        is any of the valid `classes` in gameanalysis.reduction.
    red_players : ndarray-like, optional
        The amount of players to reduce the game to. This is required for
        reductions that require them, but can be left out for identity and
        twins reductions.
    regret_thresh : float, optional
        The maximum regret to consider an equilibrium an equilibrium.
    dist_thresh : float, optional
        The minimum norm between two mixed profiles for them to be considered
        distinct.
    max_resamples : int > 0, optional
        The maximum number of times to resample a subgame when no equilibria
        can be found before giving up.
    subgame_size : int > 0, optional
        The maximum subgame support size with which beneficial deviations must
        be explored. Subgames with support larger than this are queued and only
        explored in the event that no equilibrium can be found in beneficial
        deviations smaller than this.
    num_equilibria : int > 0, optional
        The number of equilibria to attempt to find. Only one is guaranteed,
        but this might be beneifical if the game has a known degenerate
        equilibria, but one which is still helpful as a deviating strategy.
    num_backups : int > 0, optional
        In the event that no equilibrium can be found in beneficial deviations
        to small subgames, other subgames will be explored. This parameter
        indicates how many subgames for each role should be explored.
    devs_by_role : boolean, optional
        If specified, deviations will only be explored for each role in order,
        proceeding to the next role only when no beneficial deviations are
        found. This can reduce the number of profiles sampled, but may also
        fail to find certain equilibria due to the different path through
        subgames.
    at_least_one : boolean, optional
        If specified, at least one equilibria will be found in each subgame.
        This has the potential to run for a very long time as it may take
        exponential time. If your regret threshold is not set too log for your
        game, this is relatively reasonable though.
    """
    return _InnerLoop(
        prof_sched, rsgame.emptygame_copy(game), red,
        np.broadcast_to(np.asarray(red_players), game.num_roles)
        if red_players is not None else None, regret_thresh=regret_thresh,
        dist_thresh=dist_thresh, max_resamples=max_resamples,
        subgame_size=subgame_size, num_equilibria=num_equilibria,
        num_backups=num_backups, devs_by_role=devs_by_role,
        at_least_one=at_least_one).run()


class _InnerLoop(object):
    """Object to keep track of inner loop progress"""

    def __init__(self, sched, game, red, red_players, regret_thresh,
                 dist_thresh, max_resamples, subgame_size, num_equilibria,
                 num_backups, devs_by_role, at_least_one):
        # Data
        self._sched = _Scheduler(sched, game, red, red_players)
        self._game = game

        # Parameters
        self._num_eq = num_equilibria
        self._num_backups = num_backups
        self._sub_size = subgame_size
        self._max_resamples = max_resamples
        self._regret_thresh = regret_thresh
        self._dist_thresh = dist_thresh
        self._init_role_dev = 0 if devs_by_role else None
        self._at_least_one = at_least_one

        # Bookkeeping
        self._threads = queue.Queue()
        self._exp_subgames = collect.BitSet()
        self._exp_subgames_lock = threading.Lock()
        self._exp_mix = collect.MixtureSet(dist_thresh)
        self._exp_mix_lock = threading.Lock()
        self._backups = [queue.PriorityQueue() for _
                         in range(self._game.num_roles)]
        self._equilibria = []
        self._run_lock = threading.Lock()  # only one `run` at a time
        self._nash_lock = threading.Lock()  # nash is not thread safe
        self._exception = None

    def _add_subgame(self, sub_mask, count):
        if count > self._max_resamples:  # pragma: no cover
            _log.error("couldn't find equilibrium in subgame %s",
                       self._game.to_subgame_repr(sub_mask))
            return
        with self._exp_subgames_lock:
            schedule = count > 1 or self._exp_subgames.add(sub_mask)
        if schedule:
            _log.info('scheduling subgame %s%s',
                      self._game.to_subgame_repr(sub_mask),
                      ' {:d}'.format(count) if count > 1 else '')
            thread = threading.Thread(
                target=lambda: self._run_subgame(sub_mask, count))
            thread.start()
            self._threads.put(thread)

    def _run_subgame(self, sub_mask, count):
        if self._exception is not None:
            return  # Something failed
        try:
            game = self._sched.get_subgame(sub_mask, count).subgame(sub_mask)
            assert game.is_complete(), \
                "scheduled game is not complete somehow"
            if self._exception is not None:
                return  # Something failed
            with self._nash_lock:
                with warnings.catch_warnings():
                    # XXX For some reason, linesearch in optimize throws a
                    # run-time warning when things get very small negative.
                    # This is potentially a error with the way we compute
                    # gradients, but it's not reproducible, so we ignore it.
                    warnings.simplefilter('ignore', RuntimeWarning)
                    start = time.time()
                    eqa = subgame.translate(
                        nash.mixed_nash(
                            game, regret_thresh=self._regret_thresh,
                            dist_thresh=self._dist_thresh,
                            at_least_one=self._at_least_one),
                        sub_mask)
                    duration = time.time() - start
                    if duration > 600:  # pragma: no cover
                        _log.warning(
                            'equilibrium finding took %.0f seconds  in '
                            'subgame %s', duration,
                            self._game.to_subgame_repr(sub_mask))

            if eqa.size:
                for eqm in eqa:
                    self._add_deviations(eqm, self._init_role_dev)
            else:
                self._add_subgame(sub_mask, count + 1)  # pragma: no cover
        except Exception as ex:  # pragma: no cover
            self._exception = ex

    def _add_deviations(self, mix, role_index):
        with self._exp_mix_lock:
            unseen = ((role_index is not None and role_index > 0)
                      or self._exp_mix.add(mix))
        if unseen:
            _log.info(
                'scheduling deviations from %s%s', self._game.to_mix_repr(mix),
                '' if role_index is None else ' with role {}'.format(
                    self._game.role_names[role_index]))
            thread = threading.Thread(
                target=lambda: self._run_deviations(mix, role_index))
            thread.start()
            self._threads.put(thread)

    def _run_deviations(self, mix, role_index):
        if self._exception is not None:
            return  # Something failed
        try:
            support = mix > 0
            game = self._sched.get_deviations(support, 1, role_index)
            if self._exception is not None:
                return  # Something failed
            gains = regret.mixture_deviation_gains(game, mix)
            if role_index is None:
                assert not np.isnan(gains).any(), "There were nan regrets"
                if np.all(gains <= self._regret_thresh):  # Found equilibrium
                    _log.warning('found equilibrium %s with regret %f',
                                 self._game.to_mix_repr(mix),
                                 gains.max())
                    self._equilibria.append(mix)  # atomic
                else:
                    for ri, rgains in enumerate(np.split(
                            gains, game.role_starts[1:])):
                        self._queue_subgames(support, rgains, ri)

            else:  # Set role index
                rgains = np.split(gains, game.role_starts[1:])[role_index]
                assert not np.isnan(rgains).any(), "There were nan regrets"
                if np.all(rgains <= self._regret_thresh):  # No deviations
                    role_index += 1
                    if role_index < game.num_roles:  # Explore next deviation
                        self._add_deviations(mix, role_index)
                    else:  # found equilibrium
                        _log.warning('found equilibrium %s with regret %f',
                                     self._game.to_mix_repr(mix), gains.max())
                        self._equilibria.append(mix)  # atomic
                else:
                    self._queue_subgames(support, rgains, role_index)
        except Exception as ex:  # pragma: no cover
            self._exception = ex

    def _queue_subgames(self, support, role_gains, role_index):
        rs = self._game.role_starts[role_index]
        back = self._backups[role_index]

        # Handle best response
        br = np.argmax(role_gains)
        if (role_gains[br] > self._regret_thresh
                and support.sum() < self._sub_size):
            br_sub = support.copy()
            br_sub[rs + br] = True
            self._add_subgame(br_sub, 1)
        else:
            br = None  # Add best response to backup

        for si, gain in enumerate(role_gains):
            if si == br:
                continue
            sub = support.copy()
            sub[rs + si] = True
            back.put((-gain, id(sub), sub))  # id for tie-breaking

    def run(self):
        try:
            with self._run_lock:
                # Clear Data structures
                assert self._threads.empty()
                self._exp_subgames.clear()
                self._exp_mix.clear()
                for backup in self._backups:
                    while not backup.empty():
                        # This only happens when run twice in a row
                        backup.get()  # pragma: no cover
                self._equilibria.clear()
                self._thread_failed = False

                # Quiesce first time
                for sub_mask in self._game.pure_subgames():
                    self._add_subgame(sub_mask, 1)
                while not self._threads.empty():
                    self._threads.get().join()
                if self._exception is not None:
                    raise self._exception

                # Add backup games and quiesce again if less than desired
                # number of equilibria
                while (len(self._equilibria) < self._num_eq
                       and not all(q.empty() for q in self._backups)):
                    for back in self._backups:
                        for _ in range(self._num_backups):
                            if back.empty():
                                break  # pragma: no cover
                            self._add_subgame(back.get()[2], 1)
                    while not self._threads.empty():
                        self._threads.get().join()
                    if self._exception is not None:
                        raise self._exception

                # Return equilibria
                if self._equilibria:
                    return np.concatenate(
                        [eq[None] for eq in self._equilibria])
                else:
                    return np.empty((0, self._game.num_role_strats), float)  # pragma: no cover # noqa

        except Exception as ex:
            # Set exception so all threads exit
            self._exception = ex
            # Kill scheduler so all promises return
            self._sched._prof_sched.__exit__()
            # Wait for all threads to finish now that all scheduler promises
            # should have returned. This should only happen if an exception is
            # thrown internally i.e. not by the scheduler.
            while not self._threads.empty():
                self._threads.get().join()
            raise ex


class _Scheduler(object):
    """Scheduler abstraction

    This abstraction supports scheduling deviations and subgames as
    primitives"""

    def __init__(self, prof_sched, game, red, red_players):
        self._game = game
        self._red_game = red.reduce_game(game, red_players)
        self._red = red
        self._prof_sched = prof_sched
        self._profiles = {}
        self._lock = threading.Lock()

    def _get_game(self, profiles, count):
        promises = []
        with self._lock:
            for prof in profiles:
                hprof = utils.hash_array(prof)
                prof = self._profiles.setdefault(
                    hprof, _PayoffData(self._prof_sched, prof))
                promise = prof.schedule(count)
                promises.append(promise)
        payoffs = np.concatenate([prom.get()[None] for prom in promises])
        return self._red.reduce_game(paygame.game_replace(
            self._game, profiles, payoffs), self._red_game.num_role_players)

    def _subgame_profiles(self, subgame_mask):
        return self._red.expand_profiles(self._game, subgame.translate(
            self._red_game.subgame(subgame_mask).all_profiles(), subgame_mask))

    def get_subgame(self, subgame_mask, count):
        return self._get_game(self._subgame_profiles(subgame_mask), count)

    def get_deviations(self, subgame_mask, count, role_index):
        subgame_profiles = self._subgame_profiles(subgame_mask)
        dev_profiles = self._red.expand_deviation_profiles(
            self._game, subgame_mask, self._red_game.num_role_players,
            role_index)
        return self._get_game(np.concatenate((subgame_profiles, dev_profiles)),
                              count)


class _PayoffData(object):
    """Payoff data for a profile"""

    def __init__(self, prof_sched, profile):
        self._prof_sched = prof_sched
        self._profile = profile
        self._payoffs = np.zeros(profile.size, float)
        self._count = 0

        self._lock = threading.Lock()
        self._promises = collections.deque()

    def schedule(self, n):
        with self._lock:
            for _ in range(n - self._count - len(self._promises)):
                prom = self._prof_sched.schedule(self._profile)
                self._promises.append(_UpdatePromise(self, prom))
        return _PayoffPromise(self, n)


class _UpdatePromise(object):
    def __init__(self, data, promise):
        self._data = data
        self._promise = promise
        self._lock = threading.Lock()
        self._updated = False

    def wait(self):
        with self._lock:
            if not self._updated:
                # FIXME This could block on exception if schedulers don't
                # cancel all of their promises
                pay = self._promise.get()
                with self._data._lock:
                    assert self._data._promises.popleft() == self
                    self._data._count += 1
                    self._data._payoffs += (
                        (pay - self._data._payoffs) / self._data._count)
                self._updated = True


class _PayoffPromise(object):
    def __init__(self, data, n):
        self._data = data
        self._n = n

    def get(self):
        while True:
            with self._data._lock:
                assert self._n <= self._data._count + len(self._data._promises)
                if self._n <= self._data._count:
                    return self._data._payoffs
                updater = self._data._promises[0]
            updater.wait()



import itertools

import numpy as np
from gameanalysis import paygame
from gameanalysis import rsgame
from gameanalysis import utils

from egta import profsched


class SaveScheduler(profsched.Scheduler):
    """A scheduler that saves all of the payoff data for output later

    Parameters
    ----------
    game : BaseGame
        The base game of the scheduler.
    sched : Scheduler
        The bas scheduler to save payoffs from
    """

    def __init__(self, game, sched):
        self._game = rsgame.emptygame_copy(game)
        self._sched = sched
        self._payoffs = {}

    def schedule(self, profile):
        return _SavePromise(self._payoffs, profile,
                            self._sched.schedule(profile))

    def get_samplegame(self):
        by_obs = {}
        for prof, pays in self._payoffs.items():
            prof_list, pays_list = by_obs.setdefault(len(pays), ([], []))
            prof_list.append(prof.array)
            pays_list.append(pays)
        profiles = np.array(list(itertools.chain.from_iterable(
            p for p, _ in by_obs.values())), int)
        sample_pays = [np.array(p) for _, p in by_obs.values()]
        return paygame.samplegame_replace(self._game, profiles, sample_pays)

    def __enter__(self):
        self._sched.__enter__()
        return self

    def __exit__(self, *args):
        self._sched.__exit__(*args)


class _SavePromise(profsched.Promise):
    def __init__(self, data, profile, promise):
        self._data = data
        self._profile = profile
        self._promise = promise
        self._saved = False

    def get(self):
        payoff = self._promise.get()
        if not self._saved:
            payoffs = self._data.setdefault(
                utils.hash_array(self._profile), [])
            payoffs.append(payoff)
            self._saved = True
        return payoff



import logging
import queue
import threading

import numpy as np
from gameanalysis import paygame
from gameanalysis import rsgame
from gameanalysis import utils as gu

from egta import profsched
from egta import utils as eu


_log = logging.getLogger(__name__)


# TODO This should have an option where strategies with the name "noop" are
# omitted from the scheduled profile to make scheduling more efficient. This
# should be an option as it won't always be desired, and there's no easy way to
# get all of the profiles from a game.
class EgtaOnlineScheduler(profsched.Scheduler):
    """A profile scheduler that schedules through egta online

    Parameters
    ----------
    api : EgtaOnlineApi
        The api object to be uased to query EGTA Online.
    sim_id : int
        The id of the egtaonline simulator to use.
    game : RsGame
        The gameanalysis basegame representing the game to schedule.
    simultanious_obs : int
        The number of simultanious observations to schedule at a time.
        Egtaonline will use this when scheduling.
    configuration : {key: value}
        Dictionary of configuration to use with the scheduler. Any fields that
        are omitted will be filled in by the defaults specified in the
        simulator.
    sleep_time : int
        Time in seconds between queries to egtaonline to determine if profiles
        have finished. This should probably be set roughly equal to the time it
        takes for a simulation to run.
    max_scheduled : int
        The maximum number of observations to schedule simultaneously. Keeping
        this low helps prevent starving others of flux cycles.
    obs_memory : int
        The amount of memory in MB to allocate for each simulation.
    obs_time : int
        The amount of time in seconds to give each simulation to run. Too low
        and long running simulations will get cancelled giving you biased
        samples, too long and it will take longer to schedule jobs on flux.
    game_id : int, optional
        An optional game id corresponding to a game that matches the
        configuration. This prevents creating a temporary game just to query
        initial profile data.
    """

    def __init__(self, api, sim_id, game, simultanious_obs, configuration,
                 sleep_time, max_scheduled, obs_memory, obs_time,
                 game_id=None):
        self._api = api
        # XXX Copy to samplegame to get sample payoff reading
        self._game = paygame.samplegame_copy(rsgame.emptygame_copy(game))

        # {hprof: (lock, obs, _Record)}
        self._profiles = {}
        self._prof_lock = threading.Lock()
        self._num_running_profiles = 0
        # {prof_id: ^above^}
        self._prof_ids = {}
        self._runprof_lock = threading.Lock()
        self._pending_profiles = queue.Queue()

        self._sleep_time = sleep_time
        self._max_running = max_scheduled - 1
        self._sim_id = sim_id
        self._game_id = game_id
        self._configuration = configuration
        self._obs_memory = obs_memory
        self._obs_time = obs_time
        self._simult_obs = simultanious_obs

        self._sched = None
        self._running = False
        self._thread_timeout_lock = threading.Lock()
        self._exception = None

    def schedule(self, profile):
        if self._exception is not None:
            raise self._exception
        hprof = gu.hash_array(profile)
        with self._prof_lock:
            val = self._profiles.setdefault(
                hprof, (threading.Lock(), queue.Queue(), _Record()))

        _, que, _ = val
        if self._num_running_profiles > self._max_running:
            _log.debug("new profile queued %s",
                       self._game.to_prof_repr(profile))
            self._pending_profiles.put((hprof, val))
        else:
            self._schedule(hprof, val)
        return _EgtaOnlinePromise(que, self)

    def _schedule(self, hprof, val):
        """Internal implementation to actually schedule a profile

        hprof is a hashed array profile, val is the tuple stored for each
        profile. This allows differed profiles to be scheduled the same way as
        new ones."""
        profile = hprof.array
        lock, que, rec = val
        with lock:
            rec.num_req += 1

            if rec.prof_id is None:
                # Unknown id, schedule new amount
                rec.num_sched = self._simult_obs
                assignment = self._game.to_prof_repr(profile)
                _log.debug("new profile scheduled %s", assignment)
                rec.prof_id = self._sched.add_profile(
                    assignment, rec.num_sched)['id']
                self._prof_ids[rec.prof_id] = val
                with self._runprof_lock:
                    self._num_running_profiles += rec.num_sched

            elif 0 < rec.num_rec < rec.num_req and rec.num_sched == 0:
                # Id from previous data, schedule specially
                rec.num_sched = ((((rec.num_req - 1) // self._simult_obs) + 1)
                                 * self._simult_obs)
                self._prof_ids[rec.prof_id] = val
                assignment = self._game.to_prof_repr(profile)
                _log.debug(
                    "new existing data profile scheduled %s", assignment)
                self._sched.add_profile(
                    assignment, rec.num_sched)
                with self._runprof_lock:
                    self._num_running_profiles += rec.num_sched - rec.num_rec

            elif rec.num_rec < rec.num_req and rec.num_sched < rec.num_req:
                # Existing profile
                self._prof_ids[rec.prof_id] = val
                rec.num_sched += self._simult_obs
                _log.debug("old profile scheduled %s",
                           self._game.to_prof_repr(profile))
                self._sched.remove_profile(rec.prof_id)
                self._sched.add_profile(
                    self._game.to_prof_repr(profile), rec.num_sched)
                with self._runprof_lock:
                    self._num_running_profiles += self._simult_obs

    def _drain_queues(self):
        """Drain all profile queues when an assertion is thrown"""
        # This can also be thrown on keyboard interrupt
        with self._prof_lock:
            for _, que, _ in self._profiles.values():
                while que.empty():
                    que.put(None)

    def _update_counts(self):
        """Thread the constantly pings EGTA Online for updates

        This thread constantly checks an egta online scheduler for updates to
        profiles, and when found, pulls the observation data down into the
        internal structures. It's also used to schedule any profiles that
        couldn't be scheduled before because we were at max running."""
        self._thread_timeout_lock.acquire()
        try:
            while self._running:
                # First update requirements and mark completed
                _log.info("query scheduler %d", self._sched['id'])
                reqs = self._sched.get_requirements(
                )['scheduling_requirements']
                _log.debug("got reqs %s", reqs)
                for req in reqs:
                    prof_id = req['id']
                    if prof_id not in self._prof_ids:
                        continue  # race condition # pragma: no cover
                    lock, que, rec = self._prof_ids[prof_id]

                    # Check if updated counts is greater than before
                    if req['current_count'] <= rec.num_rec:
                        continue
                    _log.debug("can update profile %d", prof_id)

                    with lock:
                        # If so, get observations from profile, and update data
                        # This uses fact that recent profiles are listed first,
                        # this will not work properly if this assumption is
                        # violated
                        egta_prof = self._api.get_profile(prof_id)

                        # Sometimes egta returns invalid json
                        jobs = None
                        valid = False
                        while not valid:
                            # TODO Timeouts here are probably preferred
                            jobs = egta_prof.get_info('observations')
                            valid = all(o['symmetry_groups'] is not None for o
                                        in jobs['observations'])
                        _log.debug("obs json %s",  jobs)
                        # Parse all and slice to have accurate counts
                        new_obs = self._game.from_samplepay_json(jobs)
                        # Copy so old array can be deallocated
                        new_obs = np.copy(
                            new_obs[:new_obs.shape[0] - rec.num_rec])
                        new_obs.setflags(write=False)
                        for obs in new_obs:
                            que.put(obs)
                        finished = max(min(rec.num_sched - rec.num_rec,
                                           len(new_obs)), 0)
                        rec.num_rec += len(new_obs)
                        with self._runprof_lock:
                            self._num_running_profiles -= finished

                # Now schedule any pending profiles
                while (not self._pending_profiles.empty() and
                       self._num_running_profiles <= self._max_running):
                    hprof, val = self._pending_profiles.get()
                    self._schedule(hprof, val)

                # Now wait some time. By doing this with a lock, we allow this
                # thread to be interrupted in the case of exception.
                if self._thread_timeout_lock.acquire(True, self._sleep_time):
                    assert not self._running, \
                        "this should only be reached if we're not running"

        except Exception as ex:  # pragma: no cover
            self._exception = ex
            self._drain_queues()

        finally:
            if self._thread_timeout_lock.locked():
                self._thread_timeout_lock.release()

    def __enter__(self):
        name = 'egta_' + eu.random_string(20)

        # Create game to get initial profile data
        # TODO It would be helpful if the api had a concept of a temporary
        # game, that would be auto destroyed after context manager exit.
        gamea = None
        try:
            if self._game_id is not None:
                gamea = self._api.get_game(self._game_id)
            else:
                gamea = self._api.create_game(self._sim_id, name,
                                              self._game.num_players,
                                              self._configuration)
                _log.debug("created temp game %s %d", name, gamea['id'])
                for role, count, strats in zip(self._game.role_names,
                                               self._game.num_role_players,
                                               self._game.strat_names):
                    gamea.add_role(role, count)
                    for strat in strats:
                        gamea.add_strategy(role, strat)
            jgame = gamea.get_observations()
            assert ({(k, str(v)) for k, v in self._configuration.items()} <=
                    set(map(tuple, jgame['configuration']))), \
                "games configuration didn't match"
            sim = self._api.get_simulator(self._sim_id).get_info()
            assert ('{}-{}'.format(sim['name'], sim['version']) ==
                    jgame['simulator_fullname']), \
                "game didn't use the appropriate simulator"
            assert (
                {(r, c, frozenset(s)) for r, c, s
                 in zip(self._game.role_names, self._game.num_role_players,
                        self._game.strat_names)} ==
                {(s['name'], s['count'], frozenset(s['strategies'])) for s
                 in jgame['roles']}), \
                "game didn't have the proper role configuration"
            profiles = jgame.get('profiles', ()) or ()
        finally:
            if gamea is not None and self._game_id is None:
                gamea.destroy_game()

        # Parse profiles
        num_profs = len(profiles)
        num_pays = 0
        for jprof in profiles:
            prof, spays = self._game.from_profsamplepay_json(jprof)
            spays.setflags(write=False)
            hprof = gu.hash_array(prof)
            que = queue.Queue()
            num_pays += len(spays)
            for pay in spays:
                que.put(pay)
            val = (threading.Lock(), que, _Record(jprof['id'], len(spays)))
            self._profiles[hprof] = val
        _log.info("found %d existing profiles with %d payoffs", num_profs,
                  num_pays)

        # Create and start scheduler
        self._sched = self._api.create_generic_scheduler(
            self._sim_id, name, True, self._obs_memory,
            self._game.num_players, self._obs_time, self._simult_obs, 1,
            self._configuration)
        _log.warning(
            "created scheduler %s (%d) for running simulations: "
            "https://%s/generic_schedulers/%d", name, self._sched['id'],
            self._api.domain, self._sched['id'])
        for role, count in zip(self._game.role_names,
                               self._game.num_role_players):
            self._sched.add_role(role, count)

        self._running = True
        thread = threading.Thread(target=self._update_counts)
        thread.start()
        return self

    def __exit__(self, *args):
        self._running = False
        if self._thread_timeout_lock.locked():
            self._thread_timeout_lock.release()
        if self._sched is not None:
            self._sched.deactivate()
        self._drain_queues()
        _log.info("deactivated scheduler %d", self._sched['id'])


class _Record(object):
    """All mutable data associated with a profile"""

    def __init__(self, prof_id=None, num_rec=0):
        self.prof_id = prof_id
        self.num_req = 0
        self.num_sched = 0
        self.num_rec = num_rec


class _EgtaOnlinePromise(profsched.Promise):
    def __init__(self, que, sched):
        self._val = None
        self._queue = que
        self._sched = sched

    def get(self):
        if self._val is None:
            self._val = self._queue.get()
        if self._sched._exception is not None:
            raise self._sched._exception
        return self._val



"""Python script for performing egta"""
import argparse
import json
import logging
import smtplib
import sys
import traceback
from logging import handlers

from egtaonline import api
from gameanalysis import gamereader

from egta import countsched
from egta import savesched
from egta.script import bootstrap
from egta.script import brute
from egta.script import eosched
from egta.script import gamesched
from egta.script import innerloop
from egta.script import simsched


_log = logging.getLogger(__name__)


# TODO Create a scheduler that runs jobs on flux, but without going through
# egta online, potentially using spark
# TODO Add zip scheduler that works like simsched, but with a zip and remove
# zip wrapper
# TODO Add more efficient regret / deviation gains scheduler

def main():
    parser = argparse.ArgumentParser(
        description="""Command line egta. To run, both an equilibrium finding
        method and a profile scheduler must be specified. Each element
        processes the arguments after it, e.g. `egta -a brute -b game -c` will
        pass -a to the base parser, -b to the brute equilibrium solver, and -c
        to the game profile scheduler.""")

    # Standard arguments
    parser.add_argument(
        '--output', '-o', metavar='<output-file>', type=argparse.FileType('w'),
        default=sys.stdout, help="""The file to write the output to. (default:
        stdout)""")
    parser.add_argument(
        '--verbose', '-v', action='count', default=0, help="""Sets the
        verbosity of commands. Output is send to standard error.""")
    parser.add_argument(
        '-e', '--email_verbosity', action='count', default=0, help="""Verbosity
        level for emails.""")
    parser.add_argument(
        '-r', '--recipient', metavar='<email-address>', action='append',
        default=[], help="""Specify an email address to receive email logs at.
        Can specify multiple email addresses.""")
    parser.add_argument(
        '--count', metavar='<sims>', type=int, default=1, help="""Number of
        simulations to schedule to count towards a single payoff. (default:
        %(default)d)""")
    parser.add_argument(
        '--profile-data', metavar='<gamefile>', help="""If specified, record
        all of the payoff data sampled, and write them as an observation game
        file to the specified location.""")
    parser.add_argument(
        '--support-thresh', metavar='<support-thresh>', type=float,
        default=1e-3, help="""Threshold for a strategy probability to consider
        it in support. (default: %(default)g)""")
    parser.add_argument(
        '--tag', metavar='<identifier>', help="""This tag will be displayed in
        logs to help identify which script the log refers to. If unspecified
        but a Game ID is, then "Game <id>" will be used.""")

    # Egta online authentication
    parser_auth = parser.add_mutually_exclusive_group()
    parser_auth.add_argument(
        '--auth-string', '-a', metavar='<auth-string>', help="""The string
        authorization token to connect to egta online.""")
    parser_auth.add_argument(
        '--auth-file', '-f', metavar='<auth-file>', help="""Filename that just
        contains the string of the auth token.""")

    # Information about the game to sample
    game_spec = parser.add_mutually_exclusive_group(required=True)
    game_spec.add_argument(
        '--game-id', '-g', metavar='<game-id>', type=int, help="""An egtaonline
        game id that specifies that game profile to load. A game profile is a
        configuration, number of players per role, and strategy names per
        roll.""")
    game_spec.add_argument(
        '--game-json', metavar='<game-json>', type=argparse.FileType('r'),
        help="""A game file that must specify the number of players per role,
        and names of all role and strategies in a standard game analysis
        format.""")

    # All of the actual methods to run
    eq_methods = parser.add_subparsers(
        title='operations', dest='method', metavar='<operation>', help="""The
        operation to run on the game. Available commands are:""")
    for module in [innerloop, brute, bootstrap]:
        method = module.add_parser(eq_methods)
        method.run = module.run

        schedulers = method.add_subparsers(
            title='schedulers', dest='scheduler', metavar='<scheduler>',
            help="""The scheduler to use for acquiring samples from profiles.
            Available commands are:""")
        method.sched = schedulers
        for module in [eosched, gamesched, simsched]:
            sched = module.add_parser(schedulers)
            sched.create_scheduler = module.create_scheduler

    # Parse args and process
    args = parser.parse_args()
    method = eq_methods.choices[args.method]
    sched = method.sched.choices[args.scheduler]

    if args.auth_string is None and args.auth_file is not None:
        with open(args.auth_file) as auth_file:  # pragma: no cover
            args.auth_string = auth_file.read().strip()

    if args.tag is not None:
        tag = args.tag
    elif args.game_id is not None:
        tag = 'Game {:d}'.format(args.game_id)
    else:
        tag = '?'

    stderr_handle = logging.StreamHandler(sys.stderr)
    stderr_handle.setLevel(50 - 10 * min(args.verbose, 4))
    stderr_handle.setFormatter(logging.Formatter(
        '%(asctime)s ({}) %(levelname)s %(message)s'.format(tag)))
    log_handlers = [stderr_handle]

    # Email Logging
    if args.recipient:  # pragma: no cover
        smtp_host = 'localhost'

        # We need to do this to match the from address to the local host name
        # otherwise, email logging will not work. This seems to vary somewhat
        # by machine
        with smtplib.SMTP(smtp_host) as server:
            smtp_fromaddr = 'EGTA Online <egta_online@{host}>'.format(
                host=server.local_hostname)

        email_handler = handlers.SMTPHandler(
            smtp_host, smtp_fromaddr, args.recipient,
            'EGTA Status for {}'.format(tag))
        email_handler.setLevel(50 - args.email_verbosity * 10)
        email_handler.setFormatter(logging.Formatter(
            '%(levelname)s %(message)s'))
        log_handlers.append(email_handler)

    logging.basicConfig(level=0, handlers=log_handlers)

    try:
        if args.game_id is not None:
            with api.EgtaOnlineApi(auth_token=args.auth_string) as ea:
                jgame = ea.get_game(args.game_id).get_observations()
        else:
            jgame = json.load(args.game_json)
        game = gamereader.read(jgame)

        extra = {}
        if 'configuration' in jgame:
            extra['configuration'] = dict(jgame['configuration'])
        if 'simulator_fullname' in jgame:
            extra['simname'] = jgame['simulator_fullname']

        sched = sched.create_scheduler(game, args, **extra)
        if args.profile_data is not None:
            prof_data = savesched.SaveScheduler(game, sched)
            sched = prof_data
        if args.count > 1:
            sched = countsched.CountScheduler(sched, args.count)
        with sched:
            method.run(sched, game, args)

        if args.profile_data is not None:
            gamej = prof_data.get_samplegame().to_json()
            with open(args.profile_data, 'w') as f:
                json.dump(gamej, f)

    except KeyboardInterrupt:  # pragma: no cover
        _log.critical('execution interrupted by user')

    except Exception as ex:  # pragma: no cover
        exc_type, exc_value, exc_traceback = sys.exc_info()
        _log.critical(''.join(traceback.format_exception(
            exc_type, exc_value, exc_traceback)))
        raise ex

# Empty line to stop vim from interpreting last except as a modeline



"""An abstract profile scheduler"""
import abc


class Scheduler(abc.ABC):  # pragma: no cover
    """A profile scheduler

    It must be a context manager for its resources."""

    @abc.abstractmethod
    def schedule(self, profile):
        """Schedule a profile

        Return a promise for the payoff data"""
        pass

    @abc.abstractmethod
    def __enter__(self):
        return self

    @abc.abstractmethod
    def __exit__(self, *args):
        pass


class Promise(abc.ABC):  # pragma: no cover
    """A promise for the payoff to a profile"""

    @abc.abstractmethod
    def get(self):
        """Get the payoff data blocking until available"""
        pass



import math
import random
import threading

from gameanalysis import utils

from egta import profsched


# TODO Add common random seed for deterministic runs.


class RsGameScheduler(profsched.Scheduler):
    """Schedule profiles by adding noise to a game

    This scheduler will generate random parameters to assign to each profile.
    To generate a sample payoff, the scheduler will add noise to the payoff for
    that profile, generated as a function of the parameter. In order to work
    for compact game representations this will only sample a number of profiles
    logarithmic in the number of profiles in the complete game.

    Parameters
    ----------
    game : Game
        The game with payoff data to use. An exception will be thrown if a
        profile doesn't have any data in the game.
    noise_dist : (*params) -> ndarray, optional
        A distribution which takes the parameters for the profile and generates
        random additive payoff noise for each strategy. Strategies that aren't
        played will be zeroed out. This allows using different distributions
        for different roles.
    param_dist () -> *params, optional
        A function for generating the parameters for each profile that govern
        how payoff noise is distributed. By default there are no parameters,
        e.g. all noise comes from the same distribution.
    size_ratio : int, optional
        This won't generate unique parameters for more than a logarithmic
        number of profiles. This this the constant to be multiplied by the log
        of the game size.
    """

    def __init__(self, game, noise_dist=lambda: 0, param_dist=lambda: (),
                 size_ratio=200):
        self._noise_dist = noise_dist
        self._param_dist = param_dist
        self._max_size = max(int(math.log(game.num_all_profiles) * size_ratio),
                             game.num_profiles)
        self._game = game
        self._params = {}
        self._lock = threading.Lock()

    def schedule(self, profile):
        index = hash(utils.hash_array(profile)) % self._max_size
        with self._lock:
            params = self._params.get(index, None)
            if params is None:
                params = self._param_dist()
                self._params[index] = params
        payoff = self._game.get_payoffs(profile) + self._noise_dist(*params)
        payoff[profile == 0] = 0
        payoff.setflags(write=False)
        return _GamePromise(payoff)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass


class SampleGameScheduler(profsched.Scheduler):
    """Schedule profiles by adding noise to a sample game

    This scheduler will generate random parameters to assign to each profile.
    To generate a sample payoff, the scheduler will sample a random payoff
    associated with the profile, and then add noise generated as a function of
    the parameter.

    Parameters
    ----------
    sgame : SampleGame
        A sample game with potentially several payoffs for each profile. An
        exception will be thrown if a profile doesn't have any data in the
        game.
    noise_dist : *params -> ndarray, optional
        A distribution which takes the parameters for the profile and generates
        random additive payoff noise for each strategy. Strategies that aren't
        played will be zeroed out. This allows using different distributions
        for different roles.
    param_dist () -> *params, optional
        A function for generating the parameters for each profile that govern
        how payoff noise is distributed. By default there are no parameters,
        e.g. all noise comes from the same distribution.
    """

    def __init__(self, sgame, noise_dist=lambda: 0, param_dist=lambda: ()):
        assert hasattr(sgame, 'get_sample_payoffs'), "sgame not a sample game"
        self._noise_dist = noise_dist
        self._param_dist = param_dist
        self._sgame = sgame
        self._paymap = {}
        self._lock = threading.Lock()

    def schedule(self, profile):
        hprof = utils.hash_array(profile)
        with self._lock:
            pays, params = self._paymap.get(hprof, (None, None))
            if pays is None:
                params = self._param_dist()
                pays = self._sgame.get_sample_payoffs(profile)
                self._paymap[hprof] = (pays, params)

        pay = pays[random.randrange(pays.shape[0])]
        payoff = pay + self._noise_dist(*params)
        payoff[profile == 0] = 0
        return _GamePromise(payoff)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass


class _GamePromise(profsched.Promise):
    def __init__(self, payoff):
        self._payoff = payoff

    def get(self):
        return self._payoff



from gameanalysis import paygame

from egta import gamesched


def add_parser(subparsers):
    parser = subparsers.add_parser(
        'game', help="""Schedule simulations with noise added to a game""",
        description="""Sample profiles by adding noise to game data.""")
    parser.add_argument(
        '--sample', '-s', action='store_true', help="""Run the game scheduler
        in sample mode, where first sample payoffs are selected randomly before
        adding noise.""")
    return parser


def create_scheduler(game, args, **_):
    if args.sample:
        return gamesched.SampleGameScheduler(paygame.samplegame_copy(game))
    else:
        return gamesched.RsGameScheduler(game)



"""Script utility for running inner loop"""
import json

from gameanalysis import rsgame
from gameanalysis.reduction import deviation_preserving as dpr
from gameanalysis.reduction import identity as ir

from egta import innerloop


def add_parser(subparsers):
    parser = subparsers.add_parser(
        'quiesce', help="""Compute equilibria using the quiesce procedure""",
        description="""Samples profiles from small subgames, expanding subgame
        support by best responses to candidate subgame equilibria. For games
        with a large number of players, a reduction should be specified.""")
    parser.add_argument(
        '--dpr', metavar='<role:count,role:count,...>', help="""Specify a DPR
        reduction.""")
    parser.add_argument(
        '--regret-thresh', metavar='<reg>', type=float, default=1e-3,
        help="""Regret threshold for a mixture to be considered an equilibrium.
        (default: %(default)g)""")
    parser.add_argument(
        '--dist-thresh', metavar='<norm>', type=float, default=1e-3,
        help="""Norm threshold for two mixtures to be considered distinct.
        (default: %(default)g)""")
    parser.add_argument(
        '--max-resamples', metavar='<resamples>', type=int, default=10,
        help="""Number of times to resample all profiles in a subgame when
        equilibrium is not identified.  (default: %(default)d)""")
    parser.add_argument(
        '--max-subgame-size', metavar='<support>', type=int, default=3,
        help="""Support size threshold, beyond which subgames are not required
        to be explored.  (default: %(default)d)""")
    parser.add_argument(
        '--num-equilibria', metavar='<num>', type=int, default=1,
        help="""Number of equilibria requested to be found. This is mainly
        useful when game contains known degenerate equilibria, but those
        strategies are still useful as deviating strategies. (default:
        %(default)d)""")
    parser.add_argument(
        '--num-backups', metavar='<num>', type=int, default=1, help="""Number
        of backup subgames to pop at a time, when no equilibria are confirmed
        in initial required set.  When games get to this point they can quiesce
        slowly because this by default pops one at a time. Increasing this
        number can get games like tis to quiesce more quickly, but naturally,
        also schedules more, potentially unnecessary, simulations. (default:
        %(default)d)""")
    parser.add_argument(
        '--dev-by-role', action='store_true', help="""Explore deviations in
        role order instead of all at once. By default, when checking for
        beneficial deviations, all role deviations are scheduled at the same
        time. Setting this will check one role at a time. If a beneficial
        deviation is found, then that subgame is scheduled without exploring
        deviations from the other roles.""")
    parser.add_argument(
        '--one', action='store_true', help="""Guarantee that an equilibrium is
        found in every subgame. This may take up to exponential time, but a
        warning will be logged if it takes more than five minutes.""")
    return parser


def parse_reduction(red):
    strats = (s.strip().split(':') for s in red.strip().split(','))
    return {r.strip(): int(c) for r, c in strats}


def run(scheduler, game, args):
    game = rsgame.emptygame_copy(game)
    red = ir
    red_players = None
    if args.dpr is not None:
        red_players = game.from_role_json(parse_reduction(args.dpr), dtype=int)
        red = dpr

    eqa = innerloop.inner_loop(
        scheduler, game, red, red_players, regret_thresh=args.regret_thresh,
        dist_thresh=args.dist_thresh, max_resamples=args.max_resamples,
        subgame_size=args.max_subgame_size, num_equilibria=args.num_equilibria,
        num_backups=args.num_backups, devs_by_role=args.dev_by_role,
        at_least_one=args.one)

    for eqm in eqa:
        json.dump(game.to_mix_json(eqm), args.output)
        args.output.write('\n')



"""A scheduler that gets payoffs from a local simulation"""
import json
import logging
import queue
import subprocess
import threading

from gameanalysis import paygame
from gameanalysis import rsgame

from egta import profsched


_log = logging.getLogger(__name__)


class SimulationScheduler(profsched.Scheduler):
    """Schedule profiles using a command line program

    Parameters
    ----------
    game : RsGame
        A gameanalysis game that indicates how array profiles should be turned
        into json profiles.
    config : {key: value}
        A dictionary mapping string keys to values that will be passed to the
        simulator in the standard simulation spec format.
    command : [str]
        A list of strings that represents a command line program to run. This
        program must accept simulation spec files as flushed lines of input to
        standard in, and write the resulting output as an observation to
        standard out. After all input lines have been read, this must flush the
        output otherwise this could hang waiting for results that are trapped
        in a buffer.
    sleep : int, optional
        Time in seconds to wait before checking programs stdout for results.
        Too low and a lot of cycles will be wasted querying an empty buffer,
        too fast and programs may be waiting for results while this is
        sleeping.
    """

    def __init__(self, game, config, command, sleep=1):
        self.game = paygame.game_copy(rsgame.emptygame_copy(game))
        self.base = {'configuration': config}
        self.command = command
        self.sleep = sleep

        self._running = False
        self._proc = None
        self._inthread = None
        self._outthread = None
        self._lock = threading.Lock()
        self._inqueue = queue.Queue()
        self._outqueue = queue.Queue()

        self._exception = None

    def schedule(self, profile):
        assert self._running and self._proc is not None, \
            "can't call schedule before entering scheduler"
        if self._exception is not None:
            raise self._exception
        promise = _SimPromise(self, profile)
        self._inqueue.put(promise)
        return promise

    def _enqueue(self):
        """Thread used to push lines to stdin"""
        try:
            while self._running:
                prom = self._inqueue.get()
                if prom is None:
                    return  # told to terminate
                jprof = self.game.to_prof_json(prom._prof)
                self.base['assignment'] = jprof
                json.dump(self.base, self._proc.stdin, separators=(',', ':'))
                self._proc.stdin.write('\n')
                self._proc.stdin.flush()
                self._outqueue.put(prom)
                _log.debug("sent profile: %s",
                           self.game.to_prof_repr(prom._prof))
        except Exception as ex:  # pragma: no cover
            self._exception = ex
        finally:
            while not self._inqueue.empty():
                prom = self._inqueue.get()
                prom is None or prom._set(None)

    def _dequeue(self):
        """Thread used to get output from simulator

        This thread is constantly polling the simulator process and processing
        payoff data when its found."""
        try:
            while self._running:
                # FIXME If proc is killed, it doesn't close the streams, and as
                # a result, this call blocks indefinitely. I can't figure out a
                # way around this, but as it currently stands, this thread
                # doesn't die in that case.
                line = self._proc.stdout.readline()
                if not line and self._outqueue.empty():
                    # Process closed stdout and had nothing else to run
                    return
                elif not line or self._proc.poll() is not None:
                    # Process terminated unexpectedly
                    raise RuntimeError(
                        "Process died unexpectedly with code {}".format(
                            self._proc.poll()))
                else:
                    try:
                        jpays = json.loads(line)
                    except json.JSONDecodeError:
                        raise RuntimeError(
                            "Couldn't decode \"{}\" as json".format(line))
                    payoffs = self.game.from_payoff_json(jpays)
                    payoffs.setflags(write=False)
                    promise = self._outqueue.get()
                    if promise is None:
                        return  # told to exit
                    _log.debug("read payoff for profile: %s",
                               self.game.to_prof_repr(promise._prof))
                    promise._set(payoffs)
        except Exception as ex:  # pragma: no cover
            self._exception = ex
        finally:
            while not self._outqueue.empty():
                prom = self._outqueue.get()
                prom is None or prom._set(None)

    def __enter__(self):
        self._running = True

        self._proc = subprocess.Popen(
            self.command, stdout=subprocess.PIPE, stdin=subprocess.PIPE,
            universal_newlines=True)

        # We start these as daemons so that if we fail to close them for
        # whatever reason, python still exits
        self._inthread = threading.Thread(target=self._enqueue, daemon=True)
        self._inthread.start()
        self._outthread = threading.Thread(target=self._dequeue, daemon=True)
        self._outthread.start()
        return self

    def __exit__(self, *args):
        self._running = False

        # This tells threads to die
        self._inqueue.put(None)
        self._outqueue.put(None)

        # killing the process should close the streams and kill the threads
        if self._proc is not None and self._proc.poll() is None:
            # Kill process nicely, and then not nicely
            try:
                self._proc.terminate()
            except ProcessLookupError:  # pragma: no cover
                pass  # race condition, process died
            try:
                self._proc.wait(self.sleep)
            except subprocess.TimeoutExpired:
                _log.warning("couldn't terminate simulation, killing it...")
                self._proc.kill()
            try:
                self._proc.wait(self.sleep)
            except subprocess.TimeoutExpired:  # pragma: no cover
                _log.error("couldn't kill simulation")

        # Threads should be dead at this point, but we close anyways
        if self._inthread is not None and self._inthread.is_alive():  # pragma: no cover # noqa
            self._inthread.join(self.sleep * 2)
            if self._inthread.is_alive():
                _log.warning("couldn't kill dequeue thread...")
        if self._outthread is not None and self._outthread.is_alive():  # pragma: no cover # noqa
            self._outthread.join(self.sleep * 2)
            if self._outthread.is_alive():
                _log.warning("couldn't kill dequeue thread...")


class _SimPromise(profsched.Promise):
    def __init__(self, sched, prof):
        self._event = threading.Event()
        self._sched = sched
        self._prof = prof

    def _set(self, value):
        self._value = value
        self._event.set()

    def get(self):
        self._event.wait()
        if self._sched._exception is not None:
            raise self._sched._exception
        assert self._sched._running, \
            "can't get promise when scheduler is not running"
        return self._value



import argparse
import json

from egta import simsched


def add_parser(subparsers):
    parser = subparsers.add_parser(
        'sim', help="""Schedule simulations with a local simulator""",
        description="""Sample profiles from a local simulator specified as an
        executable.""")
    parser.add_argument(
        '--conf', '-c', metavar='<conf-json>', type=argparse.FileType('r'),
        help="""The configuration file for the simulator. This only needs to be
        specified if a configuration couldn't be found in the specified
        game.""")
    parser.add_argument(
        '--sleep', '-s', metavar='<sleep-seconds>', default=1, type=int,
        help="""Amount of time to sleep in seconds while waiting for output
        from the simulator. Too long and the script will be paused waiting for
        a response from the simulator, too short and a lof of cpu cycles will
        be wasted checking an empty buffer. (default: %(default)d)""")
    parser.add_argument(
        'command', nargs='+', metavar='<command>', help="""Command to
        execute. It must read from stdin where each line is a simulation spec
        file, and print results to stdout. stdout must be flushed after all
        inputs have been processes, otherwise this will hang waiting for a
        flushed buffer. The supplied command will be executed in the current
        directory.""")
    return parser


def create_scheduler(game, args, configuration=None, **_):
    assert configuration is not None or args.conf is not None, \
        "`conf` must be specified or supplied in the game"
    if args.conf is not None:
        configuration = json.load(args.conf)
    return simsched.SimulationScheduler(game, configuration, args.command,
                                        sleep=args.sleep)
